#include "qam.h"

#include <stdlib.h>
#include <stdint.h>
#include <stdio.h>
#include <math.h>
#include <string.h>

static inline float qam_phasediff(float a, float b) {
  float diff = a - b;
  return (diff < 0.0f) ? (
    (float)M_PI - fmodf(((float)M_PI-diff), (float)(2*M_PI))
  ) : (
    fmodf(((float)M_PI+diff), (float)(2*M_PI)) - (float)M_PI
  );
}

static void qam_softgraydecode(
  qam_demodulator_t *state, float f, float **output
) {
  float transition;
  unsigned int nibble1, nibble2;
  unsigned int mask;
  f = state->config->max_amplitude_factor * (1.0f - f / state->maxamp) / 2.0f;
  if (f < 0) f = 0;
  else if (f > state->config->max_amplitude_factor) {
    f = state->config->max_amplitude_factor;
  }
  nibble1 = floorf(f);
  nibble2 = ceilf(f);
  transition = 0.5f * (1.0f-cosf((f - nibble1) * (float)M_PI));
  nibble1 ^= nibble1 >> 1;
  nibble2 ^= nibble2 >> 1;
  for (mask=1<<(state->config->bits_per_symbol/2-1); mask; mask>>=1) {
    *(*output)++ = state->carrier_detect * (
      ((nibble1 & mask) ? -2.0f : 2.0f) * (1.0f-transition) +
      ((nibble2 & mask) ? -2.0f : 2.0f) * transition
    );
  }
}

static void qam_measure(
  qam_demodulator_t *state, float *input, float *I, float *Q, int count
) {
  int sample;
  float phase;
  *I = 0.0f; *Q = 0.0f;
  for (sample=0; sample<count; sample++) {
    phase = (
      state->phaseoffset +
      state->zerophase +
      state->config->carrier_phase_step * sample
    );
    *I += input[sample] * sinf(phase) * state->phase_odd;
    *Q += input[sample] * cosf(phase) * state->phase_odd;
  }
}

static float qam_measurephase(
  qam_demodulator_t *state, float *input, int count
) {
  float I, Q;
  qam_measure(state, input, &I, &Q, count);
  return atan2f(Q, I);
}

static void qam_measuresymbol(qam_demodulator_t *state, float *input) {
  float I, Q;
  float *output = state->output;
  qam_measure(state, input, &I, &Q, state->config->hold_samples);
  qam_softgraydecode(state, I, &output);
  qam_softgraydecode(state, Q, &output);
}

static int qam_compare_float(float *a, float *b) {
  if (*a < *b) return -1;
  if (*a > *b) return 1;
  return 0;
}

static float qam_pickmedian(qam_config_t *config, float *buffer) {
  qsort(
    buffer,
    config->qamsync_symbol_count,
    sizeof(float),
    (void *)qam_compare_float
  );
  return buffer[config->qamsync_symbol_count/2];
}

static float qam_ampmedian(qam_config_t *config, float *array) {
  float tmp[config->qamsync_symbol_count];
  memcpy(tmp, array, config->qamsync_symbol_count*sizeof(float));
  return qam_pickmedian(config, tmp);
}

static float qam_phasemedian(qam_config_t *config, float *array) {
  float tmp[config->qamsync_symbol_count], x = 0.0f, y = 0.0f, avr;
  int i;
  for (i=0; i<config->qamsync_symbol_count; i++) {
    x += cosf(array[i]);
    y += sinf(array[i]);
  }
  avr = atan2f(y, x);
  for (i=0; i<config->qamsync_symbol_count; i++) {
    tmp[i] = qam_phasediff(array[i], avr) + avr;
  }
  return qam_pickmedian(config, tmp);
}

static void qam_measureqamsync(qam_demodulator_t *state, float *input) {
  float I, Q, phase, old_zerophase;
  old_zerophase = state->zerophase;
  state->zerophase = M_PI/4;
  qam_measure(state, input, &I, &Q, state->config->hold_samples);
  state->maxamps[state->measurepos] = sqrtf((I*I + Q*Q) / 2.0f);
  state->zerophases[state->measurepos] = phase = atan2f(Q, I);
  state->maxamp = qam_ampmedian(state->config, state->maxamps);
  state->zerophase = qam_phasemedian(state->config, state->zerophases);
  if (
    qam_phasediff(phase, old_zerophase) >
    state->config->qamsync_phase_precision
  ) {
    state->carrier_detect /= 2.0f;
  } else {
    state->carrier_detect += 0.5f;
    if (state->carrier_detect > 1.0f) state->carrier_detect = 1.0f;
  }
  state->measurepos = (
    (state->measurepos + 1) % state->config->qamsync_symbol_count
  );
}

static int qam_measuresymbol_resync(qam_demodulator_t *state, float *input) {
  state->qamresync = (
    (state->qamresync + 1) % state->config->qamsync_resync_interval
  );
  if (!state->qamresync) {
    float diff;
    qam_measureqamsync(state, input);
    diff = qam_phasediff(state->zerophase, state->syncphase);
    if (diff > state->config->carrier_phase_step/2.0f) {
      state->symresync++;
    } else if (diff < -state->config->carrier_phase_step/2.0f) {
      state->symresync--;
    }
    if (state->symresync == state->config->symsync_resync_threshold) {
      state->symresync = 0;
      state->phaseoffset = fmodf(
        state->phaseoffset - state->config->carrier_phase_step,
        2.0f * (float)M_PI
      );
      state->syncphase = fmodf(
        state->syncphase + state->config->carrier_phase_step,
        2.0f * (float)M_PI
      );
      return -1;
    } else if (state->symresync == -state->config->symsync_resync_threshold) {
      state->symresync = 0;
      state->phaseoffset = fmodf(
        state->phaseoffset + state->config->carrier_phase_step,
        2.0f * (float)M_PI
      );
      state->syncphase = fmodf(
        state->syncphase - state->config->carrier_phase_step,
        2.0f * (float)M_PI
      );
      return 1;
    }
    return 0;
  } else {
    qam_measuresymbol(state, input);
    return QAM_SUCCESS;
  }
}

static int qam_measuresymsync(qam_demodulator_t *state) {
  float phaseA, phaseB, phaseC, phasediffAC, phasediffBA, phasediffBC;
  phaseA = qam_measurephase(state, state->input, state->config->hold_samples);
  state->phase_odd *= state->config->hold_odd;
  phaseB = qam_measurephase(
    state,
    state->input + state->config->hold_samples,
    state->config->switch_samples
  );
  state->phase_odd *= state->config->switch_odd;
  phaseC = qam_measurephase(
    state,
    state->input + state->config->symbol_samples,
    state->config->hold_samples
  );
  phasediffAC = fabsf(qam_phasediff(phaseA, phaseC));
  phasediffBA = fabsf(qam_phasediff(phaseB, phaseA));
  phasediffBC = fabsf(qam_phasediff(phaseB, phaseC));
  if (phasediffAC > (float)M_PI - state->config->symsync_phase_precision) {
    if (++state->symsync > state->config->symsync_rx_count) {
      state->symsync = state->config->symsync_rx_count;
    }
  } else {
    if (phasediffAC < state->config->symsync_phase_precision) {
      if (state->symsync == state->config->symsync_rx_count) {
        return QAM_SUCCESS;
      }
    }
    state->symsync -= 2;
    if (state->symsync < 0) state->symsync = 0;
  }
  if (phasediffBA < phasediffBC) return 1;
  if (phasediffBA > phasediffBC) return -1;
  return 0;
}

static void qam_demodulator_init(
  qam_demodulator_t *state,
  qam_config_t *config,
  qam_reader_fptr reader,
  void *reader_ctx
) {
  size_t incomplete_frame = state->input_bytepos % sizeof(float);
  void *tmp1 = state->output;
  void *tmp2 = state->input;
  void *tmp3 = state->maxamps;
  void *tmp4 = state->zerophases;
  if (!config->valid) abort();
  memset(state, 0, sizeof(qam_demodulator_t));
  state->output = tmp1;
  state->input = tmp2;
  state->maxamps = tmp3;
  state->zerophases = tmp4;
  state->config = config;
  state->reader = reader;
  state->reader_ctx = reader_ctx;
  state->phase_odd = 1;
  if (incomplete_frame) {
    state->input_bytepos = incomplete_frame;
    state->stage = QAM_STAGE_FRAMERECOVER;
  }
}

qam_demodulator_t *qam_demodulator_create(
  qam_config_t *config,
  qam_reader_fptr reader,
  void *reader_ctx
) {
  qam_demodulator_t *state;
  if (!config->valid) return NULL;
  state = malloc(sizeof(qam_demodulator_t));
  if (!state) return NULL;
  state->output = malloc(config->bits_per_symbol * sizeof(float));
  state->input = malloc(
    // NOTE: configuration ensures that config->init_samples is same or lower
    config->qamsync_symbol_count * config->hold_samples * sizeof(float)
  );
  state->maxamps = malloc(config->qamsync_symbol_count * sizeof(float));
  state->zerophases = malloc(config->qamsync_symbol_count * sizeof(float));
  if (state->output && state->input && state->maxamps && state->zerophases) {
    qam_demodulator_init(state, config, reader, reader_ctx);
    return state;
  } else {
    qam_demodulator_destroy(state);
    return NULL;
  }
}

void qam_demodulator_reset(qam_demodulator_t *state) {
  qam_demodulator_init(
    state,
    state->config,
    state->reader,
    state->reader_ctx
  );
}

void qam_demodulator_destroy(qam_demodulator_t *state) {
  free(state->output);
  free(state->input);
  free(state->maxamps);
  free(state->zerophases);
  free(state);
}

static int qam_readneeded(qam_demodulator_t *state, int needed) {
  needed *= sizeof(float);
  state->input_bytepos += state->reader(
    state->reader_ctx,
    (uint8_t *)state->input + state->input_bytepos,
    needed - state->input_bytepos
  );
  if (state->input_bytepos == needed) {
    state->input_bytepos = 0;
    return 1;
  }
  return 0;
}

bool qam_demodulator_run(qam_demodulator_t *state, int symsync_symbol_count) {
  int i, fails;
  while (1) {
    if (state->stage == QAM_STAGE_SUSPENDED) {
      if (!qam_readneeded(
        state,
        /* NOTE: use max buffer size for dummy read */
        state->config->qamsync_symbol_count * state->config->hold_samples
      )) return false;
      // do nothing with data
    } else if (state->stage == QAM_STAGE_FRAMERECOVER) {
      if (!qam_readneeded(state, 1)) return false;
      state->stage = QAM_STAGE_SYMSYNC;
    } else if (state->stage == QAM_STAGE_SYMSYNC) {
      if (!qam_readneeded(state, state->config->init_samples)) return false;
      if ((state->adjust = qam_measuresymsync(state)) == QAM_SUCCESS) {
        state->input_bytepos = state->config->init_samples * sizeof(float);
        state->stage = QAM_STAGE_QAMSYNC;
      } else {
        memmove(
          state->input,
          state->input + state->config->symbol_samples + state->adjust,
          (state->config->hold_samples - state->adjust) * sizeof(float)
        );
        state->input_bytepos = (
          (state->config->hold_samples - state->adjust) * sizeof(float)
        );
        if (symsync_symbol_count > 0 && !--symsync_symbol_count) return false;
      }
    } else if (state->stage == QAM_STAGE_QAMSYNC) {
      if (!qam_readneeded(
        state, state->config->qamsync_symbol_count * state->config->hold_samples
      )) return false;
      for (i=0; i<state->config->qamsync_symbol_count; i++) {
        qam_measureqamsync(state, state->input + i*state->config->hold_samples);
        state->phase_odd *= state->config->hold_odd;
      }
      fails = 0;
      for (i=0; i<state->config->qamsync_symbol_count; i++) {
        if (
          fabsf(qam_phasediff(state->zerophases[i], state->zerophase)) >
          state->config->symsync_phase_precision
        ) fails++;
      }
      if (fails > state->config->qamsync_symbol_count / 2) {
        qam_demodulator_reset(state);
      } else {
        state->phase_odd *= state->config->switch_odd;
        state->syncphase = state->zerophase;
        state->adjust = 0;
        state->carrier_detect = 1.0f;
        state->stage = QAM_STAGE_SYNCED;
      }
    } else {
      if (!qam_readneeded(
        state, state->config->symbol_samples + state->adjust
      )) return false;
      state->adjust = qam_measuresymbol_resync(
        state,
        state->input + state->config->switch_samples + state->adjust
      );
      state->phase_odd *= state->config->hold_odd * state->config->switch_odd;
      if (state->adjust == QAM_SUCCESS) {
        state->adjust = 0;
        return true;
      } else {
        return false;
      }
    }
  }
}

void qam_demodulator_suspend(qam_demodulator_t *state) {
  state->stage = QAM_STAGE_SUSPENDED;
}

bool qam_demodulator_is_suspended(qam_demodulator_t *state) {
  return state->stage == QAM_STAGE_SUSPENDED;
}

bool qam_demodulator_is_syncing(qam_demodulator_t *state) {
  return state->symsync == state->config->symsync_rx_count;
}

